##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################

EXCSAT = 0x1041
SYNCCTL = 0x1055
SYNCRSY = 0x1069
ACCSEC = 0x106D
SECCHK = 0x106E
SYNCLOG = 0x106F
ACCRDB = 0x2001
BGNBND = 0x2002
BNDSQLSTT = 0x2004
CLSQRY = 0x2005
CNTQRY = 0x2006
DRPPKG = 0x2007
DSCSQLSTT = 0x2008
ENDBND = 0x2009
EXCSQLIMM = 0x200A
EXCSQLSTT = 0x200B
EXCSQLSET = 0x2014
OPNQRY = 0x200C
PRPSQLSTT = 0x200D
RDBCMM = 0x200E
RDBRLLBCK = 0x200F
REBIND = 0x2010
DSCRDBTBL = 0x2012
SQLDTA = 0x2412
SQLDTARD = 0x2413
SQLSTT = 0x2414
SQLATTR = 0x2450
SQLSTTVRB = 0x2419
QRYDSC = 0x241A
QRYDTA = 0x241B
SQLRSLRD = 0x240E
SQLCINRD = 0x240B
ACCSECRD = 0x14AC
AGENT = 0x1403
CODPNT = 0x000C
CODPNTDR = 0x0064
CSTMBCS = 0x2435
CCSIDDBC = 0x119D
CCSIDMBC = 0x119E
CCSIDMGR = 0x14CC
UNICODEMGR = 0x1C08
CCSIDSBC = 0x119C
CMNAPPC = 0x1444
CMNSYNCPT = 0x147C
CMNTCPIP = 0x1474
XAMGR = 0x1C01
CRRTKN = 0x2135
TRGDFTRT = 0x213B
DICTIONARY = 0x1458
DEPERRCD = 0x119B
DSCERRCD = 0x2101
EXCSATRD = 0x1443
EXTNAM = 0x115E
FIXROWPRC = 0x2418
FRCFIXROW = 0x2410
LMTBLKPRC = 0x2417
MGRLVLLS = 0x1404
MGRLVLN = 0x1473
MONITOR = 0x1900
MONITORRD = 0x1C00
NEWPASSWORD = 0x11DE
PASSWORD = 0x11A1
PKGDFTCST = 0x2125
PKGID = 0x2109
MAXBLKEXT = 0x2141
MAXRSLCNT = 0x2140
RSLSETFLG = 0x2142
RDBCMTOK = 0x2105
PKGNAMCT = 0x2112
PKGSNLST = 0x2139
PRCCNVCD = 0x113F
PRDID = 0x112E
OUTOVR = 0x2415
OUTOVROPT = 0x2147
PKGCNSTKN = 0x210D
PRDDTA = 0x2104
QRYINSID = 0x215B
QRYBLKCTL = 0x2132
QRYBLKSZ = 0x2114
QRYPRCTYP = 0x2102
QRYCLSIMP = 0x215D
QRYCLSRLS = 0x215E
QRYOPTVAL = 0x215F
NBRROW = 0x213A
OUTEXP = 0x2111
PRCNAM = 0x2138
QRYATTUPD = 0x2150
RDB = 0x240F
RDBACCCL = 0x210F
RDBALWUPD = 0x211A
QRYRELSCR = 0x213C
QRYSCRORN = 0x2152
QRYROWNBR = 0x213D
QRYROWSNS = 0x2153
QRYRFRTBL = 0x213E
QRYATTSCR = 0x2149
QRYATTSNS = 0x2157
QRYBLKRST = 0x2154
QRYROWSET = 0x2156
QRYRTNDTA = 0x2155
RDBINTTKN = 0x2103
RDBNAM = 0x2110
RDBCOLID = 0x2108
RSCNAM = 0x112D
RSCTYP = 0x111F
RSNCOD = 0x1127
RSYNCMGR = 0x14C1
RTNSQLDA = 0x2116
TYPSQLDA = 0x2146
SECCHKCD = 0x11A4
SECMEC = 0x11A2
SECMGR = 0x1440
SECMGRNM = 0x1196
SECTKN = 0x11DC
RTNEXTDTA = 0x2148
SPVNAM = 0x115D
SQLAM = 0x2407
SQLCARD = 0x2408
SQLCSRHLD = 0x211F
SQLDARD = 0x2411
SRVCLSNM = 0x1147
SRVDGN = 0x1153
SRVLST = 0x244E
SRVNAM = 0x116D
SRVRLSLV = 0x115A
STTDECDEL = 0x2121
STTSTRDEL = 0x2120
SUPERVISOR = 0x143C
SVCERRNO = 0x11B4
SVRCOD = 0x1149
SYNCPTMGR = 0x14C0
SYNERRCD = 0x114A
TYPDEFNAM = 0x002F
TYPDEFOVR = 0x0035
UOWDSP = 0x2115
USRID = 0x11A0
VRSNAM = 0x1144
PKGNAMCSN = 0x2113
DIAGLVL = 0x2160
DSCINVRM = 0x220A
CMDATHRM = 0x121C
CMDCHKRM = 0x1254
CMDNSPRM = 0x1250
AGNPRMRM = 0x1232
BGNBNDRM = 0x2208
ABNUOWRM = 0x220D
ACCRDBRM = 0x2201
CMDCMPRM = 0x124B
MGRLVLRM = 0x1210
MGRDEPRM = 0x1218
ENDUOWRM = 0x220C
OBJNSPRM = 0x1253
PRCCNVRM = 0x1245
PRMNSPRM = 0x1251
PKGBNARM = 0x2206
PKGBPARM = 0x2209
QRYNOPRM = 0x2202
QRYPOPRM = 0x220F
RDBACCRM = 0x2207
SECCHKRM = 0x1219
RDBAFLRM = 0x221A
RDBATHRM = 0x22CB
RDBNACRM = 0x2204
RDBNFNRM = 0x2211
RDBUPDRM = 0x2218
RSCLMTRM = 0x1233
SYNTAXRM = 0x124C
TRGNSPRM = 0x125F
VALNSPRM = 0x1252
SQLERRRM = 0x2213
OPNQRYRM = 0x2205
ENDQRYRM = 0x220B
DTAMCHRM = 0x220E
OPNQFLRM = 0x2212
RSLSETRM = 0x2219
CMDVLTRM = 0x221D
CMMRQSRM = 0x2225
EXTDTA = 0x146C
FDODSC = 0x0010
FDODTA = 0x147A
FDODSCOFF = 0x2118
FDOPRMOFF = 0x212B
FDOTRPOFF = 0x212A
PBSD = 0xC000
PBSD_ISO = 0xC001
PBSD_SCHEMA = 0xC002
RLSCONV = 0x119F
SYNCCRD = 0x1248
XARETVAL = 0x1904
TIMEOUT = 0x1907
FORGET = 0x1186
SYNCTYPE = 0x1187
XID = 0x1801
XAFLAGS = 0x1903
RSYNCTYP = 0x11EA
SYNCRRD = 0x126D
PRPHRCLST = 0x1905
XIDCNT = 0x1906
RDBRLLBCK2 = 0xC004
DYNDTAFMT = 0x214B


SECMEC_DCESEC = 1
SECMEC_USRIDPWD = 3
SECMEC_USRIDONL = 4
SECMEC_USRIDNWPWD = 5
SECMEC_USRSBSPWD = 6
SECMEC_USRENCPWD = 7
SECMEC_USRSSBPWD = 8
SECMEC_EUSRIDPWD = 9
SECMEC_EUSRIDNWPWD = 10



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################
import socket
import ssl
import platform
import locale
import collections

from drda import codepoint as cp
from drda import ddm
from drda import secmec9
from drda import utils
from drda.cursor import Cursor


class Connection:
    def _parse_response(self):
        results = collections.deque()
        params_description = None
        description = None
        err = qrydsc = None
        chained = True
        err_msg = None
        while chained:
            dds_type, chained, number, code_point, obj = ddm.read_dds(self.sock)
            if code_point == cp.SQLERRRM:
                err_msg = ddm.parse_reply(obj).get(cp.SRVDGN).decode('utf-8')
            elif code_point == cp.SQLCARD:
                if err is None:
                    err, _ = ddm.parse_sqlcard(obj, self.encoding, self.endian)
            elif code_point == cp.SQLDARD:
                if obj[0] == 0xFF:
                    err, params_description = ddm.parse_sqldard(
                        obj, 'utf-8', self.endian, self.db_type
                    )
                else:
                    err, description = ddm.parse_sqldard(
                        obj, 'utf-8', self.endian, self.db_type
                    )
            elif code_point == cp.QRYDSC:
                ln = obj[0]
                b = obj[1:ln]
                assert b[:2] == b'\x76\xd0'
                b = b[2:]
                # [(DRDA_TYPE_xxxx, size_binary), ...]
                qrydsc = [(c[0], c[1:]) for c in [b[i:i+3] for i in range(0, len(b), 3)]]
            elif code_point == cp.QRYDTA:
                b = obj
                while len(b):
                    if (b[0], b[1]) != (0xff, 0x00):
                        break
                    b = b[2:]
                    r = []
                    for t, ps in qrydsc:
                        v, b = utils.read_field(t, ps, b, self.endian)
                        r.append(v)
                    results.append(tuple(r))
        if err:
            raise err
        return results, description, params_description

    def _parse_accsecrd(self):
        secmec = sectkn = None
        chained = True
        while chained:
            dds_type, chained, number, code_point, obj = ddm.read_dds(self.sock)
            if code_point == cp.ACCSECRD:
                while len(obj):
                    ln = int.from_bytes(obj[:2], byteorder='big')
                    sub_cp = int.from_bytes(obj[2:4], byteorder='big')
                    v = obj[4:ln]
                    obj = obj[ln:]
                    if sub_cp == cp.SECMEC:
                        secmec = int.from_bytes(v[:2], byteorder='big')
                    elif sub_cp == cp.SECTKN:
                        sectkn = v
        return secmec, sectkn

    def __init__(self, host, database, port, user, password, use_ssl, db_type):
        self.host = host
        self.database = (database + ' ' * 18)[:18]
        self.port = port
        self.user = user
        self.password = password
        self.db_type = db_type
        if self.db_type is None:
            if self.user is None:
                self.db_type = 'derby'
            elif self.user is not None:
                self.db_type = 'db2'

        if self.db_type == 'derby':
            self.encoding = 'utf-8'
            self.endian = 'big'
            self.prdid = 'DNC10130'
            self.pkgid = 'SQLC2026'
            self.pkgcnstkn = 'AAAAAfAd'
            self.pkgsn = 201
            self.user = 'APP'
            self.password = ''
            self.secmec = cp.SECMEC_USRIDONL
            self.private_key = None
        elif self.db_type == 'db2':
            self.encoding = 'cp500'
            self.endian = 'little'
            self.prdid = 'SQL11014'
            self.pkgid = 'SYSSH200'
            self.pkgcnstkn = 'SYSLVL01'
            self.pkgsn = 65
            self.user = self.user
            self.password = self.password
            # self.secmec = cp.SECMEC_USRIDPWD
            self.secmec = cp.SECMEC_EUSRIDPWD
            self.private_key = secmec9.get_private()
        else:
            raise ValueError('Unknown database type:{}'.format(self.db_type))

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if use_ssl:
            self.sock = ssl.wrap_socket(self.sock)
        self.sock.connect((self.host, self.port))

        cur_id = 1
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packEXCSAT(self, [
                cp.AGENT, 10,
                cp.SQLAM, 11,
                cp.CMNTCPIP, 5,
                cp.RDB, 12,
                cp.SECMGR, 9,
                cp.UNICODEMGR, 1208,
            ]),
            cur_id, False, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packACCSEC(
                self.database,
                self.secmec,
                secmec9.calc_public(self.private_key).to_bytes(32, byteorder='big')
                    if self.secmec == cp.SECMEC_EUSRIDPWD else None
            ),
            cur_id, False, True
        )

        secmec, sectkn = self._parse_accsecrd()

        cur_id = 1
        if secmec != self.secmec:
            self.secmec = secmec
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packACCSEC(
                    self.database,
                    self.secmec,
                    secmec9.calc_public(self.private_key).to_bytes(32, byteorder='big')
                        if self.secmec == cp.SECMEC_EUSRIDPWD else None
                ),
                cur_id, False, False
            )

        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packSECCHK(
                secmec,
                sectkn,
                self.private_key,
                self.database,
                self.user,
                self.password,
                self.encoding
            ),
            cur_id, False, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packACCRDB(self.prdid, self.database, self.encoding),
            cur_id, False, True
        )

        self._parse_response()

        self._set_valiables()

    def __enter__(self):
        return self

    def __exit__(self, exc, value, traceback):
        self.close()

    def _set_valiables(self):
        cur_id = 1
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packEXCSAT_MGRLVLLS([cp.CCSIDMGR, 1208]),
            cur_id, False, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packEXCSQLSET(self.pkgid, None, 1, self.database),
            cur_id, True, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packSQLSTT("SET CLIENT WRKSTNNAME '{}'".format(platform.node())),
            cur_id, True, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packSQLSTT("SET CURRENT LOCALE LC_CTYPE='{}'".format(locale.getlocale()[0])),
            cur_id, False, False
        )
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packRDBCMM(),
            cur_id, False, True
        )
        self._parse_response()

    def _execute(self, query, args):
        if args:
            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packPRPSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLSTT(query),
                cur_id, False, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packDSCSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, False, True
            )
            _, _, params_description = self._parse_response()

            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packEXCSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLDTA(params_description, args, self.endian),
                cur_id, False, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packRDBCMM(),
                cur_id, False, True
            )
            self._parse_response()
        else:
            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packEXCSQLIMM(
                    self.pkgid,
                    self.pkgcnstkn,
                    self.pkgsn,
                    self.database
                ),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLSTT(query),
                cur_id, False, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packRDBCMM(),
                cur_id, False, True
            )
            self._parse_response()

    def _query(self, query, args):
        if args:
            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packPRPSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLSTT(query),
                cur_id, False, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packDSCSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, False, True
            )
            _, description, params_description = self._parse_response()

            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packOPNQRY_with_params(
                    self.pkgid, self.pkgcnstkn, self.pkgsn, self.database
                ),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLDTA(params_description, args, self.endian),
                cur_id, False, True
            )
            rows, _, _ = self._parse_response()

            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packRDBCMM(),
                cur_id, False, True
            )
            _, _, _ = self._parse_response()

            return rows, description
        else:
            cur_id = 1
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packPRPSQLSTT(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, True, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packSQLSTT(query),
                cur_id, False, False
            )
            cur_id = ddm.write_request_dds(
                self.sock,
                ddm.packOPNQRY(self.pkgid, self.pkgcnstkn, self.pkgsn, self.database),
                cur_id, False, True
            )
            rows, description, params_description = self._parse_response()
            return rows, description

    def is_connect(self):
        return bool(self.sock)

    def cursor(self):
        return Cursor(self)

    def begin(self):
        self._execute("START TRANSACTION", [])

    def commit(self):
        self._execute("COMMIT", [])

    def rollback(self):
        self._execute("ROLLBACK", [])

    def close(self):
        cur_id = 1
        cur_id = ddm.write_request_dds(
            self.sock,
            ddm.packRDBCMM(),
            cur_id, False, True
        )
        self._parse_response()
        self.sock.close()



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################
# https://wiki.apache.org/db-derby/SecurityMechanism

import random

# DH Key exchange
# https://en.wikipedia.org/wiki/Diffie-Hellman_key_exchange

# p
prime = 0xC62112D73EE613F0947AB31F0F6846A1BFF5B3A4CA0D60BC1E4C7A0D8C16B3E3
# g
base = 0x4690FA1F7B9E1D4442C86C9114603FDECF071EDCEC5F626E21E256AED9EA34E4


def get_private():
    return random.randrange(2, prime)


def calc_public(private):
    return pow(base, private, prime)


def calc_session_key(public, private):
    return pow(public, private, prime).to_bytes(32, byteorder='big')


# DES/CBC/PKCS5Padding encryption
def des(server_sectkn, client_private):
    import pyDes
    assert len(server_sectkn) == 32
    # calculate session key from server_sectkn and client_private
    # get des key (8bytes) from session key

    server_public = int.from_bytes(server_sectkn, byteorder='big')
    session_key = calc_session_key(server_public, client_private)

    iv = server_sectkn[12:20]
    key = session_key[12:20]
    return pyDes.des(key, pyDes.CBC, iv, None, pyDes.PAD_PKCS5)



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################
import platform
import binascii
import drda
from drda import codepoint as cp
from drda import secmec9


def _recv_from_sock(sock, nbytes):
    n = nbytes
    recieved = b''
    while n:
        bs = sock.recv(n)
        recieved += bs
        n -= len(bs)
    return recieved


def _send_to_sock(sock, b):
    sock.send(b)


def _pack_null_string(v, enc):
    if v is None:
        return b'\xff'
    b = v.encode(enc)
    return b'\x00' + len(b).to_bytes(4, byteorder='big') + b


def _pack_binary(code_point, v):
    b = code_point.to_bytes(2, byteorder='big') + v
    return (len(b) + 2).to_bytes(2, byteorder='big') + b


def _pack_uint(code_point, v, size):
    return _pack_binary(code_point, v.to_bytes(size, byteorder='big'))


def _pack_str(code_point, v, enc):
    return _pack_binary(code_point, v.encode(enc))


def parse_string(b):
    "parse VCM"
    ln = int.from_bytes(b[:2], byteorder='big')
    if ln:
        s = b[2:2+ln].decode('utf-8')
    else:
        s = ''
    b = b[2+ln:]
    return s, b


def parse_name(b):
    "parse VCM or VCS"
    s1, b = parse_string(b)
    s2, b = parse_string(b)
    ln = int.from_bytes(b[:2], byteorder='big')
    return s1 or s2, b


def pack_dds_object(code_point, o):
    "pack to DDS packet"
    return (len(o)+4).to_bytes(2, byteorder='big') + code_point.to_bytes(2, byteorder='big') + o


def parse_reply(obj):
    d = {}
    i = 0
    while i < len(obj):
        ln = int.from_bytes(obj[i:i+2], byteorder='big')
        d[int.from_bytes(obj[i+2:i+4], byteorder='big')] = obj[i+4:i+ln]
        i += ln

    assert i == len(obj)
    return d


def parse_sqlcard(obj, enc, endian):
    if obj[0] == 0xff:
        return None, obj[1:]

    assert obj[0] == 0       # SQLCAGRP FLAG
    sqlcode = int.from_bytes(obj[1:5], byteorder=endian, signed=True)
    sqlstate = obj[5:10]
    sqlerrproc = obj[10:18]

    assert obj[18] == 0     # SQLCAXGRP FLAG
    sqlerrd = obj[19:25]
    sqlwarn = obj[25:36]

    rest = obj[36+18:]
    ln = int.from_bytes(rest[:2], byteorder='big')
    sqlrdbname = rest[2:2+ln].decode('utf-8')
    rest = rest[2+ln:]

    ln = int.from_bytes(rest[:2], byteorder='big')
    sqlerrmsg_m = rest[2:2+ln]
    rest = rest[2+ln:]

    ln = int.from_bytes(rest[:2], byteorder='big')
    sqlerrmsg_s = rest[2:2+ln]
    rest = rest[2+ln:]

    message = sqlerrmsg_m or sqlerrmsg_s

    assert rest[0] == 0xFF  # SQLDIAGGRP
    rest = rest[1:]

    if sqlcode < 0:
        err = drda.OperationalError(sqlcode, sqlstate, message)
    else:
        err = None

    return err, rest


def _parse_column_db2(b, endian, has_name):
    precision = int.from_bytes(b[:2], byteorder=endian)
    scale = int.from_bytes(b[2:4], byteorder=endian)
    sqllength = int.from_bytes(b[4:12], byteorder=endian)
    sqltype = int.from_bytes(b[12:14], byteorder=endian)
    sqlccsid = int.from_bytes(b[14:16], byteorder='big')

    b = b[16:]
    if has_name:
        b = b[6:]   # ?? skip 6 bytes
        # SQLDOPTGRP
        assert b[0] == 0x00  # not null
        b = b[3:]
        sqlname, b = parse_name(b)
        sqllabel, b = parse_name(b)
        sqlcomments, b = parse_name(b)
        b = b[7:]   # ?? skip 7 bytes
    else:
        sqllabel = None
        b = b[29:]

    return (sqllabel, sqltype, sqllength, sqllength, precision, scale, None), b


def _parse_column_derby(b, endian, has_name):
    precision = int.from_bytes(b[:2], byteorder=endian)
    scale = int.from_bytes(b[2:4], byteorder=endian)
    sqllength = int.from_bytes(b[4:12], byteorder=endian)
    sqltype = int.from_bytes(b[12:14], byteorder=endian)
    sqlccsid = int.from_bytes(b[14:16], byteorder='big')

    b = b[16:]

    # SQLDOPTGRP
    assert b[0] == 0x00  # not null
    b = b[3:]
    sqlname, b = parse_name(b)
    sqllabel, b = parse_name(b)
    sqlcomments, b = parse_name(b)

    # SQLUDTGRP
    if b[0] == 0x00:  # not null
        b = b[5:]
        sqludtrdb, b = parse_string(b)
        sqlschema, b = parse_name(b)
        sqludtname, b = parse_name(b)
    else:
        b = b[1:]

    # SQLDXGRP
    assert b[0] == 0x00  # not null
    b = b[9:]
    sqlxrdbnam, b = parse_string(b)
    sqlxcolname, b = parse_name(b)
    sqlxbasename, b = parse_name(b)
    sqlxschema, b = parse_name(b)
    sqlxname, b = parse_name(b)

    return (sqlname, sqltype, sqllength, sqllength, precision, scale, None), b


def parse_sqldard(obj, enc, endian, db_type):
    description = []
    has_name = obj[0] == 0x00
    err, rest = parse_sqlcard(obj, enc, endian)
    if not err:
        if rest[0] == 0x00:
            rest = rest[13:]
            sqlrdbnam, rest = parse_string(rest)
            sqlschema, rest = parse_name(rest)
        else:
            rest = rest[1:]
        ln = int.from_bytes(rest[0:2], byteorder=endian)
        rest = rest[2:]
        for i in range(ln):
            if db_type == 'db2':
                d, rest = _parse_column_db2(rest, endian, has_name)
            elif db_type == 'derby':
                d, rest = _parse_column_derby(rest, endian, has_name)
            description.append(d)

    return err, description


def read_dds(sock):
    "Read one DDS packet from socket"
    b = _recv_from_sock(sock, 6)
    ln = int.from_bytes(b[:2], byteorder='big')
    assert b[2] == 0xD0
    dds_type = b[3] & 0b1111
    chained = b[3] & 0b01000000
    number = int.from_bytes(b[4:6],  byteorder='big')
    obj = _recv_from_sock(sock, ln-6)

    assert int.from_bytes(obj[:2], byteorder='big') == ln - 6
    code_point = int.from_bytes(obj[2:4], byteorder='big')

    return dds_type, chained, number, code_point, obj[4:]


def write_request_dds(sock, o, cur_id, next_dds_has_same_id, last_packet):
    "Write request DDS packets"
    code_point = int.from_bytes(o[2:4], byteorder='big')
    _send_to_sock(sock, (len(o)+6).to_bytes(2, byteorder='big'))
    if code_point in (cp.SQLSTT, cp.SQLATTR, cp.SQLDTA):
        flag = 3    # DSS object
    else:
        flag = 1    # DSS request
    if not last_packet:
        flag |= 0b01000000
    if next_dds_has_same_id:
        next_id = cur_id
        flag |= 0b00010000
    else:
        next_id = cur_id + 1
    _send_to_sock(sock, bytes([0xD0, flag]))
    _send_to_sock(sock, cur_id.to_bytes(2, byteorder='big'))
    _send_to_sock(sock, o)
    cur_id = next_id
    return cur_id


def packEXCSAT(conn, mgrlvlls):
    b = b''
    for p in mgrlvlls:
        b += p.to_bytes(2, byteorder='big')

    return pack_dds_object(cp.EXCSAT, (
        _pack_str(cp.EXTNAM, 'pydrda', 'cp500') +
        _pack_str(cp.SRVNAM, platform.node(), 'cp500') +
        _pack_str(cp.SRVRLSLV, 'pydrda', 'cp500') +
        _pack_binary(cp.MGRLVLLS, b) +
        _pack_str(cp.SRVCLSNM, 'pydrda', 'cp500')
        )
    )


def packEXCSAT_MGRLVLLS(mgrlvlls):
    b = b''
    for p in mgrlvlls:
        b += p.to_bytes(2, byteorder='big')

    return pack_dds_object(cp.EXCSAT, (_pack_binary(cp.MGRLVLLS, b)))


def packSECCHK(secmec, sectkn, private_key, database, user, password, enc):
    if secmec == cp.SECMEC_EUSRIDPWD:
        des = secmec9.des(sectkn, private_key)
        return pack_dds_object(cp.SECCHK, (
                _pack_uint(cp.SECMEC, secmec, 2) +
                _pack_str(cp.RDBNAM, database, enc) +
                _pack_binary(cp.SECTKN, des.encrypt(user.encode(enc))) +
                _pack_binary(cp.SECTKN, des.encrypt(password.encode(enc)))
            )
        )
    else:
        return pack_dds_object(cp.SECCHK, (
                _pack_uint(cp.SECMEC, secmec, 2) +
                _pack_str(cp.RDBNAM, database, enc) +
                _pack_str(cp.USRID, user, enc) +
                _pack_str(cp.PASSWORD, password, enc)
            )
        )


def packACCRDB(prdid, rdbnam, enc):
    return pack_dds_object(cp.ACCRDB, (
            _pack_str(cp.RDBNAM, rdbnam, enc) +
            _pack_uint(cp.RDBACCCL, cp.SQLAM, 2) +
            _pack_str(cp.PRDID, prdid, enc) +
            _pack_str(cp.TYPDEFNAM, 'QTDSQLX86', enc) +
            _pack_binary(
                cp.CRRTKN,
                binascii.unhexlify(b'd5c6f0f0f0f0f0f12ec3f0c1f50155630d5a11')) +
            _pack_binary(
                cp.TYPDEFOVR,
                binascii.unhexlify(b'0006119c04b80006119d04b00006119e04b8'))
        )
    )


def packACCSEC(database, secmec, sectkn):
    body = (_pack_uint(cp.SECMEC, secmec, 2) +
        _pack_str(cp.RDBNAM, database, 'cp500'))
    if sectkn:
        body += _pack_binary(cp.SECTKN, sectkn)
    return pack_dds_object(cp.ACCSEC, body)


def packRDBCMM():
    return pack_dds_object(cp.RDBCMM, bytes())


def _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn):
    b = ("%-18s%-18s%-18s" % (database, "NULLID", pkgid)).encode('utf-8')
    if pkgcnstkn is None:
        b += b'\x01' * 8
    else:
        b += ("%8s" % (pkgcnstkn,)).encode('utf-8')
    return _pack_binary(
        cp.PKGNAMCSN,
        b + pkgsn.to_bytes(2, byteorder='big')
    )


def packEXCSQLSTT(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.EXCSQLSTT,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
        _pack_binary(cp.RDBCMTOK, bytes([241]))
    )


def packEXCSQLIMM(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.EXCSQLIMM,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
        _pack_binary(cp.RDBCMTOK, bytes([241]))
    )


def packPRPSQLSTT(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.PRPSQLSTT,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
        _pack_binary(cp.RTNSQLDA, bytes([241]))
    )


def packDSCSQLSTT(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.DSCSQLSTT,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
#        _pack_uint(cp.QRYINSID, 0, 8) +
        _pack_binary(cp.TYPSQLDA, bytes([1]))
    )


def packEXCSQLSET(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.EXCSQLSET,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn)
    )


def _fdodsc(description):
    _, sqltype, sqllength, _, precision, scale, _ = description
    if sqltype == 449:
        return binascii.unhexlify(b'393fff')
    elif sqltype == 485:
        return bytes([0x0f, precision, scale])
    elif sqltype == 497:
        return binascii.unhexlify(b'030004')
    else:
        raise ValueError("_fdodsc():Unknown type {}".format(sqltype))


def _fdodta(description, v):
    _, sqltype, sqllength, _, precision, scale, _ = description
    if sqltype == 449:
        v = str(v)
        return len(v).to_bytes(4, byteorder='big') + v.encode('utf_16_be')
    elif sqltype == 485:
        sign, digits, exponent = v.as_tuple()
        d = bytes([ord(b'0') + n for n in digits])
        d = (b'0' * (precision + scale) + d)[-(precision +scale):]
        d += b"d" if sign else b"c"
        if len(d) % 2:
            v = b'0' + v
        v = binascii.unhexlify(d)
        if v[0] != 0:
            v = b'\x00' + v
        return v
    elif sqltype == 497:
        v = int(v)
        return b'\x00' + v.to_bytes(4, byteorder='little')
    else:
        raise ValueError("_fdodsc():Unknown type {}".format(sqltype))


def packSQLDTA(params_desc, params, endian):
    ln = len(params)
    assert ln == len(params_desc)

    fdodsc = bytes([(1 + ln) * 3]) + binascii.unhexlify(b'76d0')
    fdodta = b''

    for i in range(ln):
        fdodsc += _fdodsc(params_desc[i])
        fdodta += _fdodta(params_desc[i], params[i])

    if (len(fdodsc) + len(fdodta)) % 2:
        fdodta = b'\x00' + fdodta

    fdodsc += binascii.unhexlify(b'0671e4d00001')

    return pack_dds_object(
        cp.SQLDTA,
        pack_dds_object(cp.FDODSC, fdodsc) +
        pack_dds_object(cp.FDODTA, fdodta)
    )


def packOPNQRY_with_params(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.OPNQRY,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
        _pack_uint(cp.QRYBLKSZ, 65535, 4) +
        _pack_uint(cp.MAXBLKEXT, 65535, 2) +
        _pack_binary(cp.QRYCLSIMP, bytes([0x01])) +
        _pack_binary(cp.DYNDTAFMT, bytes([0xf1]))
    )


def packOPNQRY(pkgid, pkgcnstkn, pkgsn, database):
    return pack_dds_object(
        cp.OPNQRY,
        _packPKGNAMCSN(database, pkgid, pkgcnstkn, pkgsn) +
        _pack_uint(cp.QRYBLKSZ, 65535, 4) +
        _pack_uint(cp.MAXBLKEXT, 65535, 2) +
        _pack_binary(cp.QRYCLSIMP, bytes([0x01]))
    )


def packSQLSTT(sql):
    return pack_dds_object(
        cp.SQLSTT,
        _pack_null_string(sql, 'utf-8') + _pack_null_string(None, 'utf-8')
    )


def packSQLATTR(attr):
    return pack_dds_object(
        cp.SQLATTR,
        _pack_null_string(attr, 'utf-8') + _pack_null_string(None, 'utf-8')
    )



import sys
import unittest
from distutils.core import setup, Command


class TestCommand(Command):
    user_options = [('db=', None, 'database type (derby|db2)')]

    def initialize_options(self):
        self.db = 'derby'

    def finalize_options(self):
        assert self.db in ('derby', 'db2'), 'Invalid database type!'

    def run(self):
        if self.db == 'derby':
            from drda.tests import test_derby as test_module
        elif self.db == 'db2':
            from drda.tests import test_db2 as test_module
        unittest.main(test_module, argv=sys.argv[:1])

cmdclass = {'test': TestCommand}

version = "%d.%d.%d" % __import__('drda').VERSION

classifiers = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Topic :: Database',
]

setup(
    name="pydrda",
    version=version,
    url='https://github.com/nakagami/pydrda/',
    classifiers=classifiers,
    keywords=['Db2', 'Apache Derby'],
    author='Hajime Nakagami',
    author_email='nakagami@gmail.com',
    description='DRDA protocol database driver',
    long_description=open('README.rst').read(),
    license="MIT",
    packages=['drda'],
    cmdclass=cmdclass,
    install_requires=['pyDes'],
)



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################

import binascii
import decimal
import datetime
import struct

DRDA_TYPE_INTEGER = 0x02
DRDA_TYPE_NINTEGER = 0x03
DRDA_TYPE_SMALL = 0x04
DRDA_TYPE_NSMALL = 0x05
DRDA_TYPE_1BYTE_INT = 0x06
DRDA_TYPE_N1BYTE_INT = 0x07
DRDA_TYPE_FLOAT16 = 0x08
DRDA_TYPE_NFLOAT16 = 0x09
DRDA_TYPE_FLOAT8 = 0x0A
DRDA_TYPE_NFLOAT8 = 0x0B
DRDA_TYPE_FLOAT4 = 0x0C
DRDA_TYPE_NFLOAT4 = 0x0D
DRDA_TYPE_DECIMAL = 0x0E
DRDA_TYPE_NDECIMAL = 0x0F
DRDA_TYPE_ZDECIMAL = 0x10
DRDA_TYPE_NZDECIMAL = 0x11
DRDA_TYPE_NUMERIC_CHAR = 0x12
DRDA_TYPE_NNUMERIC_CHAR = 0x13
DRDA_TYPE_RSET_LOC = 0x14
DRDA_TYPE_NRSET_LOC = 0x15
DRDA_TYPE_INTEGER8 = 0x16
DRDA_TYPE_NINTEGER8 = 0x17
DRDA_TYPE_LOBLOC = 0x18
DRDA_TYPE_NLOBLOC = 0x19
DRDA_TYPE_CLOBLOC = 0x1A
DRDA_TYPE_NCLOBLOC = 0x1B
DRDA_TYPE_DBCSCLOBLOC = 0x1C
DRDA_TYPE_NDBCSCLOBLOC = 0x1D
DRDA_TYPE_ROWID = 0x1E
DRDA_TYPE_NROWID = 0x1F
DRDA_TYPE_DATE = 0x20
DRDA_TYPE_NDATE = 0x21
DRDA_TYPE_TIME = 0x22
DRDA_TYPE_NTIME = 0x23
DRDA_TYPE_TIMESTAMP = 0x24
DRDA_TYPE_NTIMESTAMP = 0x25
DRDA_TYPE_FIXBYTE = 0x26
DRDA_TYPE_NFIXBYTE = 0x27
DRDA_TYPE_VARBYTE = 0x28
DRDA_TYPE_NVARBYTE = 0x29
DRDA_TYPE_LONGVARBYTE = 0x2A
DRDA_TYPE_NLONGVARBYTE = 0x2B
DRDA_TYPE_NTERMBYTE = 0x2C
DRDA_TYPE_NNTERMBYTE = 0x2D
DRDA_TYPE_CSTR = 0x2E
DRDA_TYPE_NCSTR = 0x2F
DRDA_TYPE_CHAR = 0x30
DRDA_TYPE_NCHAR = 0x31
DRDA_TYPE_VARCHAR = 0x32
DRDA_TYPE_NVARCHAR = 0x33
DRDA_TYPE_LONG = 0x34
DRDA_TYPE_NLONG = 0x35
DRDA_TYPE_GRAPHIC = 0x36
DRDA_TYPE_NGRAPHIC = 0x37
DRDA_TYPE_VARGRAPH = 0x38
DRDA_TYPE_NVARGRAPH = 0x39
DRDA_TYPE_LONGRAPH = 0x3A
DRDA_TYPE_NLONGRAPH = 0x3B
DRDA_TYPE_MIX = 0x3C
DRDA_TYPE_NMIX = 0x3D
DRDA_TYPE_VARMIX = 0x3E
DRDA_TYPE_NVARMIX = 0x3F
DRDA_TYPE_LONGMIX = 0x40
DRDA_TYPE_NLONGMIX = 0x41
DRDA_TYPE_CSTRMIX = 0x42
DRDA_TYPE_NCSTRMIX = 0x43
DRDA_TYPE_PSCLBYTE = 0x44
DRDA_TYPE_NPSCLBYTE = 0x45
DRDA_TYPE_LSTR = 0x46
DRDA_TYPE_NLSTR = 0x47
DRDA_TYPE_LSTRMIX = 0x48
DRDA_TYPE_NLSTRMIX = 0x49
DRDA_TYPE_SDATALINK = 0x4C
DRDA_TYPE_NSDATALINK = 0x4D
DRDA_TYPE_MDATALINK = 0x4E
DRDA_TYPE_NMDATALINK = 0x4F


def read_field(t, ps, b, endian):
    """
    read one field value from bytes.
    return value, rest bytes
    t: type
    ps:  precision and scale or length
    b: input bytes
    """
    if t in (
        DRDA_TYPE_NINTEGER, DRDA_TYPE_NSMALL, DRDA_TYPE_N1BYTE_INT, DRDA_TYPE_NFLOAT16,
        DRDA_TYPE_NFLOAT8, DRDA_TYPE_NFLOAT4, DRDA_TYPE_NDECIMAL, DRDA_TYPE_NNUMERIC_CHAR,
        DRDA_TYPE_NRSET_LOC, DRDA_TYPE_NINTEGER8, DRDA_TYPE_NLOBLOC, DRDA_TYPE_NCLOBLOC,
        DRDA_TYPE_NDBCSCLOBLOC, DRDA_TYPE_NROWID, DRDA_TYPE_NDATE, DRDA_TYPE_NTIME,
        DRDA_TYPE_NTIMESTAMP, DRDA_TYPE_NFIXBYTE, DRDA_TYPE_NVARBYTE, DRDA_TYPE_NLONGVARBYTE,
        DRDA_TYPE_NTERMBYTE, DRDA_TYPE_NNTERMBYTE, DRDA_TYPE_NCSTR, DRDA_TYPE_NCHAR,
        DRDA_TYPE_NVARCHAR, DRDA_TYPE_NLONG, DRDA_TYPE_NGRAPHIC, DRDA_TYPE_NVARGRAPH,
        DRDA_TYPE_NLONGRAPH, DRDA_TYPE_NMIX, DRDA_TYPE_NVARMIX, DRDA_TYPE_NLONGMIX,
        DRDA_TYPE_NCSTRMIX, DRDA_TYPE_NPSCLBYTE, DRDA_TYPE_NLSTR, DRDA_TYPE_NLSTRMIX,
        DRDA_TYPE_NSDATALINK, DRDA_TYPE_NMDATALINK,
    ):
        (isnull, b) = (b[0] == 0xFF, b[1:])
        if isnull:
            return None, b

    if t in (DRDA_TYPE_MIX, DRDA_TYPE_NMIX):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
    elif t in (DRDA_TYPE_VARMIX, DRDA_TYPE_NVARMIX):
        ln = int.from_bytes(b[:2], byteorder='big')
        v = b[2:2+ln].decode('utf-8')
        b = b[2+ln:]
    elif t == DRDA_TYPE_NVARCHAR:
        ln = int.from_bytes(b[:2], byteorder='big')
        v = b[2:2+ln].decode('utf-8')
        b = b[2+ln:]
    elif t in (
            DRDA_TYPE_SMALL, DRDA_TYPE_NSMALL, DRDA_TYPE_NINTEGER8,
            DRDA_TYPE_INTEGER, DRDA_TYPE_NINTEGER
        ):
        ln = int.from_bytes(ps, byteorder='big')
        v = int.from_bytes(b[:ln], byteorder=endian, signed=True)
        b = b[ln:]
    elif t == DRDA_TYPE_NDECIMAL:
        (p, s) = (ps[0], ps[1])
        ln = p + 1
        if ln % 2:
            ln += 1
        ln //= 2
        digits_sign = binascii.b2a_hex(b[:ln]).decode('ascii')
        sign = 0 if digits_sign[-1] == 'c' else 1
        v = decimal.Decimal(digits_sign[:-1])
        v = decimal.Decimal((sign, v.as_tuple()[1], -s))
        b = b[ln:]
    elif t in (DRDA_TYPE_TIMESTAMP, DRDA_TYPE_NTIMESTAMP):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
        v = datetime.datetime.strptime(v[:26], "%Y-%m-%d-%H.%M.%S.%f")
    elif t in (DRDA_TYPE_DATE, DRDA_TYPE_NDATE):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
        v = datetime.datetime.strptime(v, "%Y-%m-%d")
        v = datetime.date(v.year, v.month, v.day)
    elif t in (DRDA_TYPE_TIME, DRDA_TYPE_NTIME):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
        try:
            v = datetime.datetime.strptime(v, "%H:%M:%S")
        except ValueError:
            v = datetime.datetime.strptime(v, "%H.%M.%S")
        v = datetime.time(v.hour, v.minute, v.second)
    elif t in (DRDA_TYPE_VARGRAPH, DRDA_TYPE_NVARGRAPH):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
    elif t in (DRDA_TYPE_GRAPHIC, DRDA_TYPE_NGRAPHIC):
        ln = int.from_bytes(ps, byteorder='big')
        v = b[:ln].decode('utf-8')
        b = b[ln:]
    elif t in (DRDA_TYPE_NFLOAT4, ):
        ln = int.from_bytes(ps, byteorder='big')
        v = struct.unpack(">f" if endian=='big' else "<f", b[:ln])[0]
        b = b[ln:]
    elif t in (DRDA_TYPE_NFLOAT8, ):
        ln = int.from_bytes(ps, byteorder='big')
        v = struct.unpack(">d" if endian=='big' else "<d", b[:ln])[0]
        b = b[ln:]
    else:
        raise ValueError("UnknownType(%s)" % hex(t))
    return v, b


def escape_parameter(v):
    t = type(v)
    if v is None:
        return 'NULL'
    elif t == str:
        return "'" + v.replace(u"'", u"''") + "'"
    elif t == decimal.Decimal:
        return "'" + str(v) + "'"
    elif t == int or t == float:
        return str(v)
    else:
        return "'" + str(v) + "'"



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################

import datetime
import decimal
from .connection import Connection

VERSION = (0, 4, 0)
__version__ = '%s.%s.%s' % VERSION
apilevel = '2.0'
threadsafety = 1
paramstyle = 'format'


Date = datetime.date
Time = datetime.time
TimeDelta = datetime.timedelta
Timestamp = datetime.datetime


def Binary(b):
    return bytearray(b)


class DBAPITypeObject:
    def __init__(self, *values):
        self.values = values

    def __cmp__(self, other):
        if other in self.values:
            return 0
        if other < self.values:
            return 1
        else:
            return -1


STRING = DBAPITypeObject(str)
BINARY = DBAPITypeObject(bytes)
NUMBER = DBAPITypeObject(int, decimal.Decimal)
DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
DATE = DBAPITypeObject(datetime.date)
TIME = DBAPITypeObject(datetime.time)
ROWID = DBAPITypeObject()


class Error(Exception):
    def __init__(self, sqlcode, sqlstate, message):
        self.sqlcode = sqlcode
        self.sqlstate = sqlstate
        self.message = message
        super(Error, self).__init__()

    def __str__(self):
        return "%d:%s:%s" % (self.sqlcode, self.sqlstate, self.message)


class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DisconnectByPeer(Warning):
    pass


class InternalError(DatabaseError):
    def __init__(self):
        DatabaseError.__init__(self, -1, 'InternalError')


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    def __init__(self):
        DatabaseError.__init__(self, 'NotSupportedError')


def connect(host, database, port, user=None, password=None, use_ssl=False, db_type=None):
    return Connection(host, database, port, user, password, use_ssl, db_type)



##############################################################################
# The MIT License (MIT)
#
# Copyright (c) 2016-2019 Hajime Nakagami<nakagami@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################
from .utils import escape_parameter


class Cursor:
    def __init__(self, connection):
        self.connection = connection
        self.description = []
        self._rows = []
        self.arraysize = 1
        self.query = None

    def __enter__(self):
        return self

    def __exit__(self, exc, value, traceback):
        self.close()

    def callproc(self, procname, args=()):
        from drda import NotSupportedError
        raise NotSupportedError()

    def nextset(self, procname, args=()):
        from drda import NotSupportedError
        raise NotSupportedError()

    def setinputsizes(sizes):
        pass

    def setoutputsize(size, column=None):
        pass

    def execute(self, query, args=[]):
        if args and self.connection.db_type == 'derby':
            raise NotImplementedError(
                "pydrda can't execute with paramters against derby"
            )
        self.query = query
        if query.strip().split()[0].upper() == 'SELECT':
            self._rows, self.description = self.connection._query(self.query, args)
        else:
            self.connection._execute(self.query, args)

    def executemany(self, query, seq_of_params):
        rowcount = 0
        for params in seq_of_params:
            self.execute(query, params)
            rowcount += self._rowcount
        self._rowcount = rowcount

    def fetchone(self):
        from drda import OperationalError
        if not self.connection or not self.connection.is_connect():
            raise OperationalError(u"08003:Lost connection")
        if len(self._rows):
            return self._rows.popleft()
        return None

    def fetchmany(self, size=1):
        rs = []
        for i in range(size):
            r = self.fetchone()
            if not r:
                break
            rs.append(r)
        return rs

    def fetchall(self):
        r = list(self._rows)
        self._rows.clear()
        return r

    def close(self):
        self.connection = None

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def closed(self):
        return self.connection is None or not self.connection.is_connect()

    def __iter__(self):
        return self

    def __next__(self):
        r = self.fetchone()
        if not r:
            raise StopIteration()
        return r
